/*                     __                                               *\
**     ________ ___   / /  ___     Scala API                            **
**    / __/ __// _ | / /  / _ |    (c) 2003-2011, LAMP/EPFL             **
**  __\ \/ /__/ __ |/ /__/ __ |    http://scala-lang.org/               **
** /____/\___/_/ |_/____/_/ | |                                         **
**                          |/                                          **
\*                                                                      */

package scala.reflect

/**
 *  @author  Martin Odersky
 */
object NameTransformer {
  private val nops = 128
  private val ncodes = 26 * 26

  private class OpCodes(val op: Char, val code: String, val next: OpCodes)

  private val op2code = new Array[String](nops)
  private val code2op = new Array[OpCodes](ncodes)
  private def enterOp(op: Char, code: String) = {
    op2code(op) = code
    val c = (_root_.java.lang.String.instancehelper_charAt(code, 1) - 'a') * 26 + _root_.java.lang.String.instancehelper_charAt(code, 2) - 'a'
    code2op(c) = new OpCodes(op, code, code2op(c))
  }

  /* Note: decoding assumes opcodes are only ever lowercase. */
  enterOp('~', "$tilde")
  enterOp('=', "$eq")
  enterOp('<', "$less")
  enterOp('>', "$greater")
  enterOp('!', "$bang")
  enterOp('#', "$hash")
  enterOp('%', "$percent")
  enterOp('^', "$up")
  enterOp('&', "$amp")
  enterOp('|', "$bar")
  enterOp('*', "$times")
  enterOp('/', "$div")
  enterOp('+', "$plus")
  enterOp('-', "$minus")
  enterOp(':', "$colon")
  enterOp('\\', "$bslash")
  enterOp('?', "$qmark")
  enterOp('@', "$at")

  /** Replace operator symbols by corresponding `\$opname`.
   *
   *  @param name the string to encode
   *  @return     the string with all recognized opchars replaced with their encoding
   */
  def encode(name: String): String = {
    var buf: StringBuilder = null
    val len = _root_.java.lang.String.instancehelper_length(name)
    var i = 0
    while (i < len) {
      val c = _root_.java.lang.String.instancehelper_charAt(name, i)
      if (c < nops && (op2code(c) ne null)) {
        if (buf eq null) {
          buf = new StringBuilder()
          buf.append(_root_.java.lang.String.instancehelper_substring(name, 0, i))
        }
        buf.append(op2code(c))
      /* Handle glyphs that are not valid Java/JVM identifiers */
      } 
      else if (!java.lang.Character.isJavaIdentifierPart(c)) {
        if (buf eq null) {
          buf = new StringBuilder()
          buf.append(_root_.java.lang.String.instancehelper_substring(name, 0, i))
        }
        val tmp : String = {
          val h = java.lang.Integer.toHexString(c.toInt)
          "$u" + "000".take(4 - h.size) + h
        } 
        buf.append(tmp)
      }
      else if (buf ne null) {
        buf.append(c)
      }
      i += 1
    }
    if (buf eq null) name else buf.ToString
  }

  /** Replace `\$opname` by corresponding operator symbol.
   *
   *  @param name0 the string to decode
   *  @return      the string with all recognized operator symbol encodings replaced with their name
   */
  def decode(name0: String): String = {
    //System.out.println("decode: " + name);//DEBUG
    val name = if (_root_.java.lang.String.instancehelper_endsWith(name0, "<init>")) _root_.java.lang.String.instancehelper_substring(name0, 0, _root_.java.lang.String.instancehelper_length(name0) - (_root_.java.lang.String.instancehelper_length("<init>"))) + "this"
               else name0;
    var buf: StringBuilder = null
    val len = _root_.java.lang.String.instancehelper_length(name)
    var i = 0
    while (i < len) {
      var ops: OpCodes = null
      var unicode = false
      val c = _root_.java.lang.String.instancehelper_charAt(name, i)
      if (c == '$' && i + 2 < len) {
        val ch1 = _root_.java.lang.String.instancehelper_charAt(name, i+1)
        if ('a' <= ch1 && ch1 <= 'z') {
          val ch2 = _root_.java.lang.String.instancehelper_charAt(name, i+2)
          if ('a' <= ch2 && ch2 <= 'z') {
            ops = code2op((ch1 - 'a') * 26 + ch2 - 'a')
            while ((ops ne null) && !_root_.java.lang.String.instancehelper_startsWith(name, ops.code, i)) ops = ops.next
            if (ops ne null) {
              if (buf eq null) {
                buf = new StringBuilder()
                buf.append(_root_.java.lang.String.instancehelper_substring(name, 0, i))
              }
              buf.append(ops.op)
              i += _root_.java.lang.String.instancehelper_length(ops.code)
            }
            /* Handle the decoding of Unicode glyphs that are 
             * not valid Java/JVM identifiers */
          } else if ((len - i) >= 6 && // Check that there are enough characters left
                     ch1 == 'u' && 
                     ((java.lang.Character.isDigit(ch2)) || 
                     ('A' <= ch2 && ch2 <= 'F'))) {
            /* Skip past "$u", next four should be hexadecimal */
            val hex = _root_.java.lang.String.instancehelper_substring(name, i+2, i+6)
            try {
              val str = java.lang.Integer.parseInt(hex, 16).toChar
              if (buf eq null) {
                buf = new StringBuilder()
                buf.append(_root_.java.lang.String.instancehelper_substring(name, 0, i))
              }
              buf.append(str)
              /* 2 for "$u", 4 for hexadecimal number */
              i += 6
              unicode = true
            } catch {
              case _:NumberFormatException =>
                /* <code>hex</code> did not decode to a hexadecimal number, so
                 * do nothing. */
            }
                       }
        }
      }
      /* If we didn't see an opcode or encoded Unicode glyph, and the
        buffer is non-empty, write the current character and advance
         one */ 
      if ((ops eq null) && !unicode) { 
        if (buf ne null)
          buf.append(c) 
        i += 1 
      } 
    }
    //System.out.println("= " + (if (buf == null) name else buf.toString()));//DEBUG
    if (buf eq null) name else buf.ToString
  }
}
